#!/usr/bin/python3
# -*- coding: utf-8 -*-
'''
________________________________________________________________________
 a_js.py                                                        
 This file is free software under         
     the terms of the GPL without any warranty - see the file COPYING 
 Diese Datei ist Freie Software gemäß
             der GPL ohne jegliche Garantie - siehe die Datei COPYING 
________________________________________________________________________

 a_js.py stellt Tools für die anderen absenzen-Module zur Verfügung
  Thema: Rechnen mit und Konfigurieren von Jahrgangsstufen
________________________________________________________________________

Regel: js ist die Jahrgangsstufe, am bay. Gym. derzeit 1-10 für Klassen
          intern sind Klassen von 1-19 möglich
          20 ist der Oberbegriff für Kurssystem (z.B. für die User-Rechte)
          21 ist das erste Kursjahr usw. bis 25 für großzügige 5 Kursjahre
        Die Namen der Kursjahre sind konfigurierbar (Kursjahr_1 usw.)
        ks sind die Klassenstufen - typischerweise   1: Grundschule 1-4
                                                    2: Unterstufe  5-7
                                                    3: Mittelstufe 8-10
                                                    4: Kurssystem
        Sie Namen der ks sind konfigurierbar (Klassenstufe_1 usw.)
        Die Grenzen der Klassenstufen sind konfigurierbar als Beginn der zweiten und dritten Klassenstufe
            (Klassenstufe_2_Beginn usw.) da der Beginn der ersten bei 1 und das Ende der dritten bei 19 liegt:
        Typisch: 1 - ks=1 - 4 | 5 - ks=2 - 7 | 8 - ks=3 - 19 | 20 - ks=4 - unendlich
        Da die ks keine Bedeutung haben, können sie zur Strukturierung der Verwaltungsrechte herangezogen werden
'''
##Importe
from a_konstanten import *
import a_db
import a_datum


class A_KS:
    def __init__(self,beginn,ende,name):
        self.beginn=beginn
        self.ende=ende
        self.name=name

class A_KJ:
    def __init__(self,name):
        self.name=name
    
    def hole_praefix(self,nr):
        '''Holt das Praefix - das Kursjahr-identifizierende Präfix 
            Kein KJ-Name: leeres Praefix - wird also beim Einlesen nicht berücksichtigt.
            Kein Praefix: Nummer wird als Praefix gesetzt '''
        if self.name:
            param='Kursjahr_'+str(nr)+'-Praefix'
            self.praefix=a_db.get_config(param)
            if not self.praefix:
                self.praefix=str(nr)
                a_db.set_config(param,self.praefix)
        else:
            self.praefix=""
            
class A_JS:
    def __init__(self,ks,name,mindestdatum,besetzt=True):
        self.ks=ks
        self.name=name
        self.mindestdatum=mindestdatum
        self.besetzt=besetzt
    
_js_=[]
_ks_=[]
_kj_=[]

def init(): # erst aufrufen, wenn db bereits initialisiert!
    global _js_,_ks_,_kj_,_last_kj_
    
    # einlesen und logische Korrektheit sicherstellen
    ks2b=a_db.get_config_int("Klassenstufe_2_Beginn",-1)
    if ks2b<0:
        ks2b=5
        a_db.set_config("Klassenstufe_2_Beginn",ks2b)
    if ks2b<2:
        ks2b=2
        a_db.set_config("Klassenstufe_2_Beginn",ks2b)
    if ks2b>18:
        ks2b=18
        a_db.set_config("Klassenstufe_2_Beginn",ks2b)
    ks3b=a_db.get_config_int("Klassenstufe_3_Beginn",-1)
    if ks3b<0:
        ks3b=8
        a_db.set_config("Klassenstufe_3_Beginn",ks3b)
    if ks3b<=ks2b:
        ks3b=ks2b+1
        a_db.set_config("Klassenstufe_3_Beginn",ks3b)
    if ks3b>19:
        ks3b=19
        a_db.set_config("Klassenstufe_3_Beginn",ks3b)

    _kj_=[A_KJ("Kurse")]
    _last_kj_=0
    for i in range(1,6):
        kjn=a_db.get_config('Kursjahr_'+str(i))
        _kj_.append(A_KJ(kjn))
        if kjn:
            _last_kj_=20+i
        
    if not a_db.get_config("Klassenstufe_1"): a_db.set_config("Klassenstufe_1","G")
    if not a_db.get_config("Klassenstufe_2"): a_db.set_config("Klassenstufe_2","U")
    if not a_db.get_config("Klassenstufe_3"): a_db.set_config("Klassenstufe_3","M")
    # Ziel: ks[2].beginn liefert die Jahrgangsstufe, mit der die zweite Klassenstufe beginnt
    _ks_=[        A_KS(   0,    25,"alle")] # alle
    _ks_.append(  A_KS(   1,ks2b-1,a_db.get_config("Klassenstufe_1"))) # Primar
    _ks_.append(  A_KS(ks2b,ks3b-1,a_db.get_config("Klassenstufe_2"))) # Sek1
    _ks_.append(  A_KS(ks3b,    19,a_db.get_config("Klassenstufe_3"))) # Sek2
    _ks_.append(  A_KS(  20,    25,"Kurse"                              )) # Kurssystem - für die Rechte
    for i in range(5,20):
            _ks_.append(  A_KS(-1,-1,"Dummy"))
    _ks_.append(  A_KS(  21,    25,"Kurse"                              )) # Kurssystem - zum Adressieren
    for i in range(21,26):
            _ks_.append(  A_KS(i,i,_kj_[i-20].name))    

    _js_=[A_JS(0,"Alle",'2000-01-01')]
    ksptr=1
    mindestdatum=a_db.get_config('intern_mindestdatum-gum')
    for i in range(1,26):
        if i<=20:
            if i>_ks_[ksptr].ende:
                ksptr+=1
            _js_.append(A_JS(ksptr,'JS '+str(i),mindestdatum))
        else:
            mindestdatum=a_db.get_config('intern_mindestdatum-kj_'+str(i-20),'2000-01-01')
            if not a_datum.ist_isodatum(mindestdatum):
                mindestdatum='2000-01-01'
            _js_.append(A_JS(i,_kj_[i-20].name,mindestdatum))

def get_last_kj(test_kj=99): # liefert entweder das letzte Kursjahr (typisch:22) oder das übergebene, wenn dieses kleiner ist
    return min(_last_kj_,test_kj)
    
def get_ks(nr):
    return _ks_[nr]

def get_ks_intervall(nr):
    myks=_ks_[nr]
    return myks.beginn,myks.ende

def get_js(nr):
    return _js_[nr]
    
def get_kj(nr):    
    if nr>=20:
        nr-=20
    return _kj_[nr]
   
def hole_kj_kriterien():
    for i in range(1,6):
        _kj_[i].hole_praefix(i)
        
def finde_kj(name):
    for i in range(1,6):
        if _kj_[i].name==name:
            return 20+i
    return 0
