#!/usr/bin/python3
# -*- coding: utf-8 -*-
'''
________________________________________________________________________
 a_um_einlesen.py                                                            
 This file is free software under              
     the terms of the GPL without any warranty - see the file COPYING 
 Diese Datei ist Freie Software gemäß
             der GPL ohne jegliche Garantie - siehe die Datei COPYING 
________________________________________________________________________

 a_um_einlesen.py liest 
  1. die Liste (aus WinSV) aller 
    Schüler der Unter- und Mittelstufe mit ihrer Klasse ein
    - legt Klassen an
    - legt Schüler an
  2. die Liste (aus WinLD) aller
    Lehrer der Unter- und Mittelstufe ein und ordnet sie ihren
    Klassen zu
________________________________________________________________________
'''

import sys
import codecs
import time, datetime
import random

import cgi
import cgitb; cgitb.enable()
import os, sys

import hashlib
import a_main, a_tool, a_upload
import a_db
import a_js, a_datum

from a_konstanten import *

class A_Spaltenanalyse:
    def __init__(self):
        self.klasse=-1
        self.name=-1
        self.vorname=-1
        self.nachname=-1
        self.omm=-1 # ordnungsmerkmal
        
    def get_info(self):
        return "Spaltennummern: Klasse="+str(self.klasse)+" Name="+str(self.name)+" Vorname="+str(self.vorname)+" Nachname="+str(self.nachname)+" Ordnungsmerkmal="+str(self.omm)
    def get_name(self,array):
        if self.nachname>=0 and self.vorname>=0: # normal z.B. bei ASV
            return array[self.nachname]+" "+array[self.vorname]
        if self.name>=0:                         # normal z.B. bei WinSV
            return array[self.name]
        if self.nachname>=0:                     # Oops
            return array[self.nachname]
        if self.vorname>=0:                      # Oops
            return array[self.vorname]
        return 'Keine Namensspalte definiert'    # darf nie passieren!
        
def schueler_import():
    global db_conn
    global db_cursor
    importerfolg=False
    alle_zeilen=[] # wir lesen einfach alle Zeilen ein und sehen, ob es einen Fehler gibt...
    a_tool.importlog("Start der Einlesens von schuelerliste.txt: "+a_datum.get_jetzt())
    try: # klappt als utf-8 einlesen?
        with codecs.open(a_tool.get_datenpfad()+"schuelerliste.txt",mode="r",encoding='utf-8') as f:
            for zeile in f:
                alle_zeilen.append(zeile)
            a_tool.importlog("Erfolgreiches Einlesen als utf-8 ("+str(len(alle_zeilen))+" Zeilen)")
    except:
        alle_zeilen=[] 
    if len(alle_zeilen)==0:   
        try: # ok, klappt als iso-8859-1?
            with codecs.open(a_tool.get_datenpfad()+"schuelerliste.txt",mode="r",encoding='iso-8859-1') as f:
                for zeile in f:
                    alle_zeilen.append(zeile)
                a_tool.importlog("Erfolgreiches Einlesen als iso-8859-1 ("+str(len(alle_zeilen))+" Zeilen)")
        except:
            alle_zeilen=[] 
    if len(alle_zeilen)==0:
        a_tool.importlog("Keine Daten lesbar", err="Keine Daten lesbar - Datei leer oder keine Kodierung klappt")
        return False
    spaltenanalyse,neue_zeilen=analysiere_schuelerzeilen(alle_zeilen)
    omm_anzahl=a_db.get_first_element('SELECT count(ordnungsmerkmal) FROM personen')
    if spaltenanalyse.omm<0 and omm_anzahl>0:
        a_tool.importlog("",err="/Warnung!\n  |Es wurde kein 64-stelliges Ordnungsmerkmal in den einzulesenden Daten erkannt.\n  |In den Stammdaten sind aber {} Einträge mit Ordnungsmerkmal vorhanden.\n  |Es kann unerwartet zu falschen Zuordnungen bei doppelten Namen kommen!\n  |Es werden aber keine Schüler überschrieben.\n  \\Lesen Sie erneut MIT Ordnungsmerkmal ein um sicher zu gehen.".format(omm_anzahl))
    a_db.execute('UPDATE personen SET klasse="" WHERE js<19 ',do_commit=False)         # alle Klasseneinträge der UM-Schüler löschen
    zeilenzahl=0
    letzteklasse=""
    ordnungsmerkmal="" # bleibt es auch, wenn omm nicht gesetzt
    js=0
    for zeile in neue_zeilen:
        zeilenzahl+=1
        if debuglevel>0:
            a_tool.importlog("Zeilennr."+str(zeilenzahl)+": "+str(zeile))
            if debuglevel>2: a_tool.importlog("Zeile war",alle_zeilen[zeilenzahl-1]) # muss nicht sein...
        
        ignoriere=False
        for ignorierer in ignoriererliste:
            if zeile[0].startswith(ignorierer):
                ignoriere=True
        if ignoriere:
            continue
        schueler=''
        klasse=zeile[spaltenanalyse.klasse]
        schueler=spaltenanalyse.get_name(zeile) # wegen Fallunterscheidungen
        if spaltenanalyse.omm>=0:
            ordnungsmerkmal=zeile[spaltenanalyse.omm]
            
        if schueler:        
            if klasse!=letzteklasse:
                js=lege_klasse_an(klasse) # erkennt natürlich eine bestehende...
                letzteklasse=klasse
            if debuglevel>0: 
                a_tool.importlog("Schüler:"+schueler+" Klasse:"+klasse+" Jahrgangsstufe:"+str(js))
            
            logtxt,errtxt,pid=a_tool.schueler_eintragen(schueler,"",js,klasse,do_commit=False,ordnungsmerkmal=ordnungsmerkmal)      
            if logtxt or errtxt: a_tool.importlog(logtxt,err=errtxt)
            importerfolg=True
        else:
            a_tool.importlog(zeile,err='Zeile wurde nicht als Schülereintrag erkannt: "'+zeile.strip()+'"')
    f.close()
    a_db.commit()
    return importerfolg

def analysiere_schuelerzeilen(alle_zeilen):
    spaltenanalyse= A_Spaltenanalyse()
    trennerliste=[ [0,"|"],[0,";"],[0,","], ] # Elemente: [Trefferzahl,Trennzeichen] - Komma ist etwas fragwürdig
    for zeile in alle_zeilen:
        for i in range(len(trennerliste)):
            trennerliste[i][0]+=zeile.count(trennerliste[i][1])
    trennerliste.sort() 
    trenner=trennerliste[-1][1]
    a_tool.importlog("Trennzeichen für die Spalten",trenner)
    bestandteilslaengen=[] # nimmt [Längensumme,Spaltennummer] auf -> liefert sortiert die Nummern der größten Spalten
    erste_zeile=True
    kill_laufende_nummer=0
    neue_zeilen=[]
    for zeile in alle_zeilen:
        bestandteile=zeile.split(trenner)
        for i in range(min(len(bestandteile),50)): # nur die ersten 50 Spalten werden analysiert, Rest wird ignoriert, da wohl Fehler
            if i==len(bestandteilslaengen):
                bestandteilslaengen.append([0,i])
            b=bestandteile[i].strip()
            if b.startswith('"') and b.endswith('"') and len(b)>2:
                b=b[1:-1]
            if i==0 and b.isdigit():
                b='' # erste Spalte mit reinen Zahlen: fortlaufende Nummer oder anderes unnötiges
                kill_laufende_nummer+=1
            bestandteile[i]=b
            bestandteilslaengen[i][0]+=len(b)
        if erste_zeile \
            and ( ('Nachname' in bestandteile and 'Vorname' in bestandteile) or 'Name' in bestandteile) \
            and ('Klassenbezeichnung' in bestandteile or 'Klasse' in bestandteile):
            for i in range(len(bestandteile)):
                if bestandteile[i]=="Nachname":     spaltenanalyse.nachname=i
                if bestandteile[i]=="Vorname":      spaltenanalyse.vorname=i
                if bestandteile[i] in ["Klassenbezeichnung", "Klasse"]:
                                                    spaltenanalyse.klasse=i
                if bestandteile[i]=="Ordnungsmerkmal":     spaltenanalyse.omm=i
                if bestandteile[i]=="Name":         spaltenanalyse.name=i
            erste_zeile=False    
            if debuglevel:a_tool.importlog("Spaltenanalyse durch erste Zeile ok",spaltenanalyse.get_info())        
        else:
            neue_zeilen.append(bestandteile)
        
    bestandteilslaengen_backup=list(bestandteilslaengen)
    a_tool.importlog(str(bestandteilslaengen))

    ln_delta=1 if (kill_laufende_nummer>len(neue_zeilen)-3) else 0  # also die meisten ersten Zellen als laufende Nummer identifiziert -> einrechnen
    bestandteilslaengen.sort()
    if debuglevel>2:a_tool.importlog(str(bestandteilslaengen))
    maximum=-1
    if spaltenanalyse.omm<0 and (bestandteilslaengen[maximum][0]/len(neue_zeilen) )>=60: # Nur das Ordnungsmerkmal scheint so lang zu sein
        spaltenanalyse.omm=bestandteilslaengen[maximum][1]
        maximum-=1
    if erste_zeile: # es gab also keinen Spalten-Kopf! Dann alle überzähligen Spalten beseitigen! Berücksichtige dazu 2 + ggf. laufende-nummern-spalte + ggf. omm-spalte
        maxspalten=2+ln_delta+ (1 if spaltenanalyse.omm>=0 else 0)
        bestandteilslaengen=bestandteilslaengen_backup[:maxspalten]
        bestandteilslaengen.sort()
        if debuglevel>2:a_tool.importlog(str(bestandteilslaengen))
        maximum=-1 # quasi Neustart
        if spaltenanalyse.omm<0 and (bestandteilslaengen[maximum][0]/len(neue_zeilen) )>=60: # Nur das Ordnungsmerkmal scheint so lang zu sein
            # spaltenanalyse.omm=bestandteilslaengen[maximum][1] sollte ja kein neues Ergebnis bringen
            maximum-=1 # also überspringen, falls es noch drin ist
    if spaltenanalyse.name<0 and spaltenanalyse.nachname<0:
        spaltenanalyse.name=bestandteilslaengen[maximum][1]  # die sonst breiteste Spalte sind die Namen - es darf also keine Spalte "Schulname" o.ä. geben!
    if spaltenanalyse.klasse<0:
        spaltenanalyse.klasse=bestandteilslaengen[maximum-1][1]  # die nächstbreiteste Spalte sind die Klassen
    if debuglevel:a_tool.importlog("Endgültige Spaltenanalyse",spaltenanalyse.get_info())        
    return spaltenanalyse,neue_zeilen
    
def lege_klasse_an(klasse,klassleiter=""):
    js=10 # was keine ziffer enthält kommt pauschal in die 10. Klasse, also Mittelstufe
    kj=a_js.finde_kj(klasse)
    if kj: # legt keine Klasse an, liefert aber korrektes js für ein Kursjahr (21-25) zurück
        if debuglevel>0: 
            a_tool.importlog("Kursjahr gefunden -> js="+str(kj))
        return kj
    else:
        for i in range(len(klasse)):
            if len(klasse)>1+i and klasse[i] in ["0","1"] and klasse[i+1].isdigit(): # schlägt bei 00 bis 19 an
                js=int(klasse[i:i+2])
                if debuglevel>0: 
                    a_tool.importlog("js zweistellig")
                break
            elif len(klasse)>i and klasse[i].isdigit():
                js=int(klasse[i])
                if debuglevel>0: 
                    a_tool.importlog("js einstellig")
                break
    if debuglevel>0: 
        a_tool.importlog("js ermittelt aus "+klasse,str(js))
    if js<1 or js>19: # zur Sicherheit...
        js=10
    klassleiter_nr=a_tool.get_user_nr(klassleiter)
    rowdict=a_db.get_one_rowdict('''SELECT name,klassleiter_nr FROM klassen WHERE name=?''',(klasse,))
    if rowdict:
        db_name,db_klassleiter_nr = rowdict
        if klassleiter_nr and db_klassleiter_nr!=klassleiter_nr:
            a_db.execute('''UPDATE klassen SET js=?, klassleiter_nr=? WHERE name=?''',(js,klassleiter_nr,klasse),do_commit=False)
    else:
        a_db.execute('''INSERT INTO klassen (name,js,klassleiter_nr) VALUES (?,?,?)''',(klasse,js,klassleiter_nr),do_commit=False)
    if js>0 and js<a_db.get_config_int('minimum-js',99):
        a_db.set_config('minimum-js',js,do_commit=False)
    return js
    
def lehrer_import():
    global db_conn
    global db_cursor
    importerfolg=False
    alle_zeilen=[] # wir lesen einfach alle Zeilen ein und sehen, ob es einen Fehler gibt...
    a_tool.importlog("Start der Einlesens von lehrerliste.txt: "+a_datum.get_jetzt())
    try: # klappt als utf-8 einlesen?
        with codecs.open(a_tool.get_datenpfad()+"lehrerliste.txt",mode="r",encoding='utf-8') as f:
            for zeile in f:
                alle_zeilen.append(zeile)
            a_tool.importlog("Erfolgreiches Einlesen als utf-8 ("+str(len(alle_zeilen))+" Zeilen)")
    except:
        alle_zeilen=[] 
    if len(alle_zeilen)==0:   
        try: # ok, klappt als iso-8859-1?
            with codecs.open(a_tool.get_datenpfad()+"lehrerliste.txt",mode="r",encoding='iso-8859-1') as f:
                for zeile in f:
                    alle_zeilen.append(zeile)
                a_tool.importlog("Erfolgreiches Einlesen als iso-8859-1 ("+str(len(alle_zeilen))+" Zeilen)")
        except:
            alle_zeilen=[] 
    if len(alle_zeilen)==0:
        a_tool.importlog("Keine Daten lesbar", err="Keine Daten lesbar - Datei leer oder keine Kodierung klappt")
        return False

    a_db.execute('DELETE from lehrerinklasse',do_commit=False)                 # alle Lehrer-Klassen-Beziehungen löschen
    #finde trenner
    trennerliste=[ [0,"|"],[0,";"] ] # Elemente: [Trefferzahl,Trennzeichen] - kein Komma, da bei ASV Aufzählungen mit Komma getrennt!
    for zeile in alle_zeilen:
        for i in range(len(trennerliste)):
            trennerliste[i][0]+=zeile.count(trennerliste[i][1])
    trennerliste.sort() 
    if trennerliste[-1][0]>len(alle_zeilen)-5: # jede Zeile sollte wenigstens einen trenner enthalten
        trenner=trennerliste[-1][1]
    else:
        trenner=',' # Fallback, falls wirklich nur mit Kommas separiert
    #ermittle Existenz von führender laufender Nummer (als 12 oder "12")
    zaehler=0
    for zeile in alle_zeilen:
        item=zeile.split(trenner)[0].strip()
        if item.isdigit() or ( len(item)>2 and item[1:-1].isdigit() ):
            zaehler+=1
    if zaehler>len(alle_zeilen)-5:
        lehrerspalte=1
    else:
        lehrerspalte=0
    #lese endlich jede zeile aus
    zeilenzahl=0
    for zeile in alle_zeilen:
        zeilenzahl+=1
        if debuglevel>1: a_tool.importlog("Zeilennr."+str(zeilenzahl)+": "+str(zeile))
        ignoriere=False
        for ignorierer in ignoriererliste:
            if zeile.startswith(ignorierer):
                ignoriere=True
        if ignoriere:
            a_tool.importlog('ignoriere:"'+zeile+'"')
            continue
        zeilenarray=zeile.split(trenner)
        if len(zeilenarray)-lehrerspalte<2:
            a_tool.importlog('',err="Zeile zu kurz, ignoriere:'"+zeile+"'")
            continue
        for i in range(len(zeilenarray)):
            zeilenarray[i]=zeilenarray[i].replace('"','').strip()
        lehrer=zeilenarray[lehrerspalte]
        klassen=(','.join(zeilenarray[lehrerspalte+1:]) ).split(',') # alle mit "trenner" getrennten Klassen mit Komma vereint und dann am Komma gesplittet
                                                                     # Dadurch wird [hansi,"5a,5b"] genauso erkannt wie [hansi,"5a","5b"]
        if debuglevel>1: a_tool.importlog('L:{} K:{}'.format(str(lehrer),str(klassen)))
        lehrer=a_tool.bereinige_name(lehrer)
        if len(lehrer)>1:
            lehrer=a_tool.entferne_titel(lehrer)
            lehrer_nr=a_tool.lege_user_an(lehrer,do_commit=False)
            letzte_klasse=""
            for klasse in klassen:
                klasse=klasse.strip()
                if klasse in ["|",",",";","\n",""]:
                    continue # überspringen
                if "_" in klasse:
                    klasse=klasse.split("_")[0]
                if klasse!=letzte_klasse:
                    letzte_klasse=klasse
                    item=a_db.get_first_element('SELECT name FROM klassen WHERE name=?  COLLATE NOCASE',(klasse,)) # ignoriert gross/kleinschrift
                    if item:
                        klasse=item # scheint unnötig, setzt aber den richtigen namen mit gross/kleinschrift
                    else:
                        a_tool.importlog("Klasse:"+klasse+' beim Lehrereinlesen neu angelegt')
                        lege_klasse_an(klasse)
                    if not a_db.get_existenz('SELECT user_nr,kname FROM lehrerinklasse WHERE user_nr=? AND kname=?  COLLATE NOCASE',(lehrer_nr,klasse)):
                        a_db.execute('INSERT INTO lehrerinklasse (user_nr,kname) VALUES (?,?)',(lehrer_nr,klasse),do_commit=False)
                    if debuglevel>1: 
                        a_tool.importlog("Lehrer:"+lehrer+" in Klasse:"+klasse)
                    importerfolg=True
        else:
            a_tool.importlog('',err="Keine Verknüpfungen für Lehrer:"+lehrer+", da der Name zu kurz ist")
    f.close()
    a_db.commit()
    return importerfolg

            
debuglevel=2

bincgi=False

if __name__ == '__main__':
    bincgi=True
    importerfolg=False
    if a_tool.getcgiparameter('uploadtype')=='schuelerliste':
        # 1. hochgeladene Datei in Empfang nehmen und speichern
        erfolg, message=a_upload.save_uploaded_file('schuelerliste','schuelerliste.txt')
        a_tool.importlog("Schuelerlisteupload",message)
        # 2. Datei in die Datenbank einlesen
        if erfolg:  
           if True: #try:
                if schueler_import():
                    a_tool.importlog('\n\nAnfangsbuchstabenverteilung berechnen','\n'+a_tool.bestimme_anfangsbuchstabenverteilung())
                    a_tool.importlogfinish()
                    importerfolg=True
                    a_main.absentmain('+Einlesen der Schülerliste erfolgreich',kurspflicht='StammdatenS',fehlerliste=a_tool.get_importerrors())
            #except:
            #   pass
    elif a_tool.getcgiparameter('uploadtype')=='lehrerliste':
        # 1. hochgeladene Datei in Empfang nehmen und speichern
        erfolg, message=a_upload.save_uploaded_file('lehrerliste','lehrerliste.txt')
        a_tool.importlog("Lehrerlisteupload",message)
        # 2. Datei in die Datenbank einlesen
        if erfolg:  
            if True: #try:
                if lehrer_import():
                    a_tool.importlogfinish()
                    importerfolg=True
                    a_main.absentmain('+Einlesen der Lehrerliste erfolgreich',kurspflicht='StammdatenL',fehlerliste=a_tool.get_importerrors())
            #except:
            #   pass
        
    if not importerfolg:
        a_tool.importlogfinish()
        a_main.absentmain('-Fehler beim Einlesen ('+(['Nichts eingelesen?']+a_tool.logliste)[-1]+')')
