#!/usr/bin/python3
# -*- coding: utf-8 -*-
'''
________________________________________________________________________
 mail2ab.py                                                         
 This file is free software under          
     the terms of the GPL without any warranty - see the file COPYING 
 Diese Datei ist Freie Software gemäß
     der GPL ohne jegliche Garantie - siehe die Datei COPYING 
________________________________________________________________________

 mail2ab.py liest von einem konfigurierten Mailpostfach per POP3 Mails
  - enthält der Betreff einen Code der Form 123456A, so wird
    - in der db nach einer Person mit dem Code gesucht
    - dieser Person für den aktuellen Tag eine Krankmeldung eingetragen
    - und eine Bestätigungsmail geschickt
  - alles wird für den Verwalter protokolliert
________________________________________________________________________
'''

import sys, os
import codecs
import time, datetime

import smtplib, poplib, cgi
import random
from email.parser import Parser, BytesParser
from email.header import decode_header

import a_tool
import a_db, a_h

LEVEL_KRANKMELDUNG=3
timestamp=time.localtime()
jetzt=time.strftime("%Y-%m-%d %H:%M", timestamp)
heute=time.strftime("%d.%m.%y", timestamp)
dumpname=time.strftime("%Y-%m-%d_%H-%M_", timestamp)
akzeptierbar=True
if timestamp.tm_hour>=a_db.get_config_int('aa-stundemorgen',16): # nach 16 Uhr gelten Krankmeldungen für morgen!
    timestamp=time.localtime(time.mktime((timestamp.tm_year,timestamp.tm_mon,timestamp.tm_mday+1,0,0,0,0,0,0)))
elif timestamp.tm_hour>=a_db.get_config_int('aa-stundegrenze',12): # ab 12 Uhr werden Krankmeldungen abgelehnt
    akzeptierbar=False
datum=time.strftime("%Y-%m-%d", timestamp)
dodebug=False   # darf nie hier auf True gesetzt werden, sondern nur im Mail-Bereich unten (darf also wenn mail2ab als Modul genutzt wird nie True sein -> bricht cgi-Logik
diemail=b''
m2ab_erfolg='Fehler'
extrainfo='''\n
----------------------------------------------------------------------- 
 _Erklaerung_
 Nur bei Erfolg steht ein Datum hinter "Resultat:" es ist das Datum, 
            fuer das die Krankmeldung registriert wurde.
 (neu)      am Ende der Zeile bedeutet, dass die Absenz hiermit neu
            angelegt wurde (du warst schnell genug)
 (ergaenzt) bedeutet, dass du bereits vermisst wurdest.
            Die Krankmeldung wurde trotzdem eingetragen.\n

 _Fehlermeldungen_
 Ein Fehler bedeutet, dass deine Krankmeldung _nicht_ registriert wurde!
 Fehler "kein Code erkannt": 
      Stand dein Code im Betreff?
      Hatte er die Form ACD123456X ?
 Fehler "Code konnte nicht zugeordnet werden": 
      Es wurde ein Code gefunden, aber er gehoert niemandem!
      Vermutlich ein Schreibfehler!
 Fehler "konnte nicht mehr eingetragen werden": 
      Du warst schon krank gemeldet!

 Denk daran, die noetige Entschuldigung innerhalb von drei Tagen vorzulegen.
 Gute Besserung!\n
-----------------------------------------------------------------------\n
 Bitte nicht auf diese Mail antworten!
 Emails duerfen nur vor 14 Uhr (Krankmeldung fuer den aktuellen Tag)
              oder nach 17 Uhr (Krankmeldung fuer den folgenden Tag) 
              verschickt werden.
 Sollte die Rueckmeldung einmal nicht kommen oder nicht das gewuenschte
              Ergebnis anzeigen, so ist eine Meldung an das Oberstufenbuero
              hilfreich.\n\n'''
__event__=0

def set_erfolg(wert):
    global m2ab_erfolg
    m2ab_erfolg=wert
    if skriptaufruf:
        a_db.set_config('intern_last_mail2ab',jetzt+' ('+m2ab_erfolg+')')
    else:
        a_db.set_config('intern_last_mail2ab_test',jetzt+' ('+m2ab_erfolg+')')
    
def get_jetzt():
    global __event__
    __event__+=1
    return jetzt # +'_'+('00'+str(__event__)[:3])
    
def vermerke_meldung(code,adresse,adressevoll):     
    if "Testmail" in code:
        protokolliere(1,adresse,adressevoll,'absenzenautomat',code+" angekommen. Wuerde als Absenz fuer "+datum+" gelten.")
        return
        
    row=a_db.get_one_rowdict('''
        SELECT pid, name 
        FROM personen
        WHERE zugriffscode=?''',(code,))
    if row:
        pid, name=row
        print(pid,name)
        status=-1 # Auch ein Flag für Ergebnis der folgenden Abfrage
        attestplicht=0
        neuerkommentar=heute+' Krankmeldung per Mail ('+adressevoll.replace('<br>','')+')'
        zeile=a_db.get_one_rowdict('''
            SELECT status, kommentar FROM absenzen WHERE pid=? AND datum=? ''',(pid,datum))
        if zeile:
            status,kommentar=zeile
        if status==-1: #neu anlegen, da vorheriges Ergebnis leer
            a_db.execute('''
                INSERT INTO absenzen (pid, status, attestpflicht, datum, kommentar) 
                    VALUES (?,?,?,?,?) ''',(pid,LEVEL_KRANKMELDUNG,0,datum,neuerkommentar.replace(' (','<br>&nbsp;(')))
            protokolliere(1,adresse,adressevoll,name,neuerkommentar+' (neu)')
        elif status<LEVEL_KRANKMELDUNG: #Krankmeldung eintragen
            a_db.execute('''
                UPDATE absenzen
                SET status=?, kommentar=? 
                WHERE pid=? AND datum=? ''',(LEVEL_KRANKMELDUNG,kommentar+'<br>'+neuerkommentar.replace(' (','<br>&nbsp;('),pid,datum))
            protokolliere(1,adresse,adressevoll,name,neuerkommentar+' (ergaenzt)')
        else:
            protokolliere(0,adresse,adressevoll,name,'Krankmeldung konnte nicht mehr eingetragen werden')
    else:
        protokolliere(0,adresse,adressevoll,'','CODE konnte niemandem zugeordnet werden - Versuch es nochmal mit dem korrigierten Code')
        protokolliere(2,a_db.get_config('aa-adminadresse'),'admin','unbekannt','von '+adressevoll+': CODE konnte niemandem zugeordnet werden')

def testmailsenden():
    s1=protokolliere(3,a_db.get_config('aa-mailadresse'),'absenzenautomat-system','Beauftragt','Testmail von '+get_jetzt())
    s2=protokolliere(2,a_db.get_config('aa-adminadresse'),'absenzenautomat-system','Beauftragt','Testmail von '+get_jetzt()+' verschickt')
    if s1 and s2:
        return '''<h2>Testmail wurde versendet</h2><br>
            Die Ergebnisse sind jetzt und in Kürze im Maillog zu erkennen.<br>
            Sollte auch nach über 5 Minuten keine weitere Meldung auftauchen, so ist der cron-job zu überprüfen (Systemadministrator!).<br>
            Sie können mit dem entsprechenden Knopf testweise die Mails abholen lassen, um die grundsätzliche Funktion zu testen.<br>
            <b>Vorsicht:</b> Bei echten Problemen und echten Daten werden die Mails damit zum falschen Zeitpunkt eingelesen!<br>
            <br>Sie können dieses Fenster / diesen Tab nun schließen.
        </body>'''
    tm="" if s1 else 'Testmail an '+a_db.get_config('aa-mailadresse')+' konnte nicht am SMTP-Server '+a_db.get_config('aa-smtpserver')+' abgeliefert werden!<br>'
    am="" if s2 else 'Testmailbestätigung an '+a_db.get_config('aa-adminadresse')+' konnte nicht am SMTP-Server '+a_db.get_config('aa-smtpserver')+' abgeliefert werden!<br>'
    a_h.merke_body_extra('style="color:red;"')
    return '''<h2>Testmails sollten verschickt werden</h2>
        <div class="b1"><b>Es traten Fehler auf!</b><br>
        '''+tm+am+'<br>Weitere Meldungen sind jetzt und in Kürze im Maillog zu erkennen</div></body>'
    
def testmailholen():
    text='''<h2>Testweise werden Mails abgerufen</h2><br>
    Sollte dieses Abholen funktionieren, Sie aber ansonsten keine Mails erhalten,<br>
    so ist der Aufruf von /opt/absenzen/cgi-bin/mail2ab.py über /etc/crontab nicht funktionsfähig.<br>
    Es folgt der Ablauf:<br><br><div class="b1">'''
    return text+hole_mails()+'</div><div style="clear:left;"></div><br>Sie können dieses Fenster / diesen Tab nun schließen.</body>'
    
def protokolliere(erfolg,adresse,adressevoll,name,aktion):
    if dodebug: print("Ich protokolliere: "+get_jetzt()+', '+str(erfolg)+', '+adresse+', '+adressevoll+', '+name+', '+aktion)
    erfolgsmeldung=['Rueckmeldung wegen eines Fehlers','Erfolgsmeldung','Statusmeldung','Testmail','Ablehnung wg. Uhrzeit']
    success=False
    if name:
        name+=': '
    if erfolg in [0,1,3,4]:
        a_db.execute('''
            INSERT INTO poplog (timestamp,erfolg,adresse,aktion) 
                VALUES (?,?,?,?) ''',(get_jetzt(),erfolg,adressevoll,name+aktion))
        
    
    login=a_db.get_config('aa-mailadresse')
    absender=a_db.get_config('aa-adminadresse')
    try:
        smtpserver = smtplib.SMTP(a_db.get_config('aa-smtpserver'))
        smtpserver.ehlo()
        smtpserver.starttls()
        smtpserver.login(login, a_db.get_config('aa-passwort'))
        betreff='Subject:'
        if erfolg!=3:
            betreff+='Krankmeldung - '+erfolgsmeldung[erfolg]
        else:
            betreff+=erfolgsmeldung[erfolg] +' von '+ get_jetzt()
        header = 'To:' + adresse + '\n' + 'From: ' + absender + '\n' + betreff +' \n'
        msg = '\n Mail wurde verarbeitet:'+jetzt+'\n Resultat:'+aktion+'\n'+extrainfo
        #print(msg)
        msg=header +cgi.escape(msg).encode('ascii', 'xmlcharrefreplace').decode('utf-8')
        #print(msg)
        smtpserver.sendmail(absender, adresse, msg)
        success=True
        if dodebug: print('done!')
    except Exception as e:
        errortext=erklaerbaer('SMTP',str(type(e))[1:-1])
        
        if dodebug: print('Abbruch des Sendens wegen: '+errortext)
        a_db.execute('''
            INSERT INTO poplog (timestamp,erfolg,adresse,aktion) 
                VALUES (?,?,?,?) ''',(get_jetzt(),0,adressevoll,errortext.replace('\n','<br>')))
    try:
        smtpserver.close()
    except:
        pass
    return success
    
        
def dumpe(name,mailtext):
    f=open(a_tool.basepath+'cgi-bin/data/dump/'+name,"w")
    for i in range(len(mailtext)):
        line=mailtext[i:i+1]
        try:
            f.write(line.decode())
        except:
            f.write('_')
    f.close()
        
def un_quote(line):
    zeile=''
    zeilensequenz=decode_header(line)
    for item,coding in zeilensequenz:
        if type(item)== bytes:
            if coding==None:
                zeile+=item.decode('ascii')
            else:
                zeile+=item.decode(coding)
        else:
            zeile+=item
    return zeile

def analysiere_mail(mailobjekt):
    global diemail
    code=''
    adresseroh=''
    diemail=b''
    # Mail einlesen
    for line in mailobjekt:
        diemail+=line+b'\n' 
        if len(diemail)>10240: 
            diemail+=b'\n...und weitere Zeilen. Abgeschnitten da groesser 10kB                                   '
            break
    # Mail analysieren
    headers= BytesParser().parsebytes(diemail,headersonly=True)
    # Werte auslesen
    adressevoll=un_quote(str(headers['from']))
    subject=un_quote(str(headers['subject']))
    
    # Code im Betreff finden
    position=subject.find('ACD')
    if position<0:  position=subject.find('KGA') # Legacy...
    if position>=0 and len(subject)>=position+10:
        substring=subject[position+3:position+6+3]
        zeichen=subject[position+6+3]
        if dodebug: print(substring,'-',zeichen)
        if substring.isdigit():
            code=substring+zeichen      # mit Buchstabe am Ende
            print("Code gefunden! ",code)
    
    position=subject.find('Testmail')
    if position==0:
        code=subject
    # Emailadresse in der from-angabe finden
    pos1=adressevoll.find("<")+1
    pos2=adressevoll.find(">")
    if pos1>0 and pos2>pos1:
        adresse=adressevoll[pos1:pos2].strip().lower()
    else:
        adresse=adressevoll.strip().lower()
        adresse=adresse.replace("'","")
    if dodebug: print('Analyse: '+adressevoll+' schickte '+subject+ ' [reine Adresse & Code: '+adresse+' & '+code+']')
    adressevoll=adressevoll.replace('<','\n').replace('>','').replace('\n','<br>')
    return code,adresse,adressevoll

def erklaerbaer(typ,fehler):
    meldung='Fehler beim Zugriff auf den '+typ+'-Server: '
    if 'gaierror' in fehler.lower():
        meldung+='Server nicht gefunden. => Servername überprüfen!'
    if 'authenticationerror' in fehler.lower() or 'error_proto' in fehler.lower() :
        meldung+='Anmeldung fehlgeschlagen. => Benutzername und Passwort kontrollieren!'
    return meldung+'\nOriginalmeldung: '+fehler
        

def hole_mails():
    ablauf=[]
    try:
        M=poplib.POP3_SSL(  a_db.get_config('aa-popserver'))
        M.user(             a_db.get_config('aa-mailadresse'))
        M.pass_(            a_db.get_config('aa-passwort'))
    except Exception as e:
        errortext=erklaerbaer('POP',str(type(e))[1:-1])
        ablauf.append('Abbruch des Maildownloads wegen: '+errortext)
        a_db.execute('''
            INSERT INTO poplog (timestamp,erfolg,adresse,aktion) 
                VALUES (?,?,?,?) ''',(get_jetzt(),0,'absenzenautomat-system',errortext.replace('\n','<br>')))
        set_erfolg('kein POP möglich')
        a_h.merke_body_extra('style="color:red;"')
        return 'Fehler beim Mailabruf'
    numMessages=len(M.list()[1])
    ablauf.append("Anzahl der Nachrichten: "+str(numMessages))
    set_erfolg('POP-Zugriff ok')
    for m_nr in range(numMessages):
        code,adresse,adressevoll=analysiere_mail(M.retr(m_nr+1)[1])
        # Entscheidungen treffen            
        if code and adresse:
            ablauf.append(adresse+' meldet '+code)
            if akzeptierbar:
                vermerke_meldung(code,adresse,adressevoll)
            else:
                if code.find("Testmail")==0:
                    vermerke_meldung(code,adresse,adressevoll)
                nummer=dumpname+str(random.randint(100000,900000))
                protokolliere(4,adresse,adressevoll,'','Die Mail wurde nicht rechtzeitig fuer heute oder zu frueh fuer morgen geschickt')
                dumpe(nummer,diemail)
            set_erfolg('Code gefunden')
        elif code:
            ablauf.append("Unbekannt meldet "+code)
            if akzeptierbar:
                vermerke_meldung(code,"unbekannt@example.com","unbekannt")
            else:
                nummer=dumpname+str(random.randint(100000,900000))
                protokolliere(4,'','','','Die Mail wurde nicht rechtzeitig fuer heute oder zu frueh fuer morgen geschickt')
                dumpe(nummer,diemail)
            set_erfolg('Code gefunden')
        elif adresse:
            ablauf.append(adresse+' suchte erfolglos Kontakt')
            protokolliere(0,adresse,adressevoll,'unbekannt','Es wurde kein Code erkannt - Neuer Versuch mit Code ACD123456X im Betreff!')
            if not akzeptierbar:
                protokolliere(4,adresse,adressevoll,'','Die Mail wurde nicht rechtzeitig fuer heute oder zu frueh fuer morgen geschickt')
            nummer=dumpname+str(random.randint(100000,900000))
            protokolliere(2,a_db.get_config('aa-adminadresse'),'admin','Info an biec:','Mail '+nummer+' von '+adressevoll+': Es wurde kein Code erkannt')
            dumpe(nummer,diemail)
        else:
            nummer=dumpname+str(random.randint(100000,900000))
            ablauf.append("Mail ohne Absendeadresse")
            protokolliere(2,a_db.get_config('aa-adminadresse'),'admin','Info an biec:','Mail '+nummer+' ohne Absendeadresse '+adressevoll)
            dumpe(nummer,diemail)
        
        M.dele(m_nr+1)
    M.quit()
    if dodebug: print("\n".join(ablauf))
    return "<br>".join(ablauf)


# Start des Programms bei direktem Aufruf   
skriptaufruf=False
if __name__ == '__main__':
    skriptaufruf=True
    print('da bin ich: '+a_tool.basepath+'cgi-bin/')
    import getpass
    username= getpass.getuser()
    if username!="www-data":
        print('Das Skript wurde vom User '+username+' aufgerufen.\nEs sollte nur von www-data aufgerufen werden.\nTipp: Vorher mit su www-data den User wechseln, dann ggf. bash aufrufen\nEigentliches Ziel: Mit dem User www-data von cron aufrufen lassen!\n')
    dodebug=True
    hole_mails()
    
    
    
